program gen_sel
! a simple program for genomic selection; 
!analyzes data from program snp_sim
!Ignacy Misztal, Sept 08 - Oct 8, 2008
implicit none

integer::   n_snp,n_qtl,nrec,nprog
integer,parameter::r8=selected_real_kind(15)
		   
real::var_ratio	!variance ratio
integer,parameter::nround=110	!maximum number of rounds of Gauss-Seidel iteration

real::mean(3),covar(3,3),var(3),w(3),corr1,corr2,corr3
integer::i,k,l,neq
real::bv,bv1,z,z0			  
real,allocatable:: snp_val(:),snp_freq(:),xy(:),sol(:)
integer,allocatable::genome(:,:)
real,allocatable::x(:,:),xpx(:),y(:)

print '("Program GEN_SEL 1.1 (I. Misztal, UGA)")'
print*,'variance ratio?'
read*,var_ratio
call cpu_time(z0)

call read_setup
neq=n_snp+1
allocate(x(nrec,neq),xy(neq),sol(neq),xpx(neq),y(nrec))
xy=0
open(2,file='SNP_data')		! contains phenotypes + genotypes

do i=1,nrec
   read(2,'(i6,1x,2f8.2,1x,20000i1)')k,y(i),bv,genome
   x(i,:)=(/1,sum(genome,dim=1)-3/)
   xy=xy+x(i,:)*y(i)
enddo

print*,'read'

! create diagonals (without 
do i=1,neq
   xpx(i)=dot_product(x(:,i),x(:,i))
enddo

print*,'diags'
print '(" 5 elements of XPX: ",10(2x,f0.1))',xpx(1:5)

call cpu_time(z)
Print '(/"Iteration started at ",f8.2," seconds")',z-z0
!call solve_GSRU 
call solve_pcg
!calculate predictions and correlations of predicted and simulated BV

open(3,file='SNP_results')
write(3,'(" Estimated mean: ",f10.2)')sol(1)
write(3,'("Simulated and estimated genomic values")')
write(3,'(i6,2f10.2)')(i,snp_val(i),sol(i+1),i=1,n_snp)
write(3,'(//"Simulated and estimated breeding values")')

!calculate predictions + statistics
rewind 2

do l=1,2
mean=0;covar=0
if (l==1) then
     print '(//"Results for training data set")'
   else
     print '(//"Results for independent data set")'
endif
	 
do i=1,nrec
   read(2,'(i6,1x,2f8.2,1x,20000i1)')k,y(i),bv,genome
	  bv1=sum( (genome(1,:)+genome(2,:)-3) *sol(2:))
	  write(3,'(i6,2f10.2)')i,bv,bv1 
	  w=(/y(i),bv,bv1/)
	  mean=mean+w
	  covar=covar+outer_prod(w,w)
enddo

mean=mean/nrec
do i=1,3
   var(i)=(covar(i,i)-nrec*mean(i)**2)/nrec
enddo

corr1=(covar(1,3)/nrec-mean(1)*mean(3)) / sqrt(var(1))/sqrt(var(3)) 
corr2=(covar(2,3)/nrec-mean(2)*mean(3)) / sqrt(var(2))/sqrt(var(3))
corr3=(covar(1,2)/nrec-mean(1)*mean(2)) / sqrt(var(1))/sqrt(var(2))
print '(20x,"y           simulated_bv     genomic_bv")'
print '(" mean",t10,3f15.2)',mean
print '(" var",t10,3f15.2)',var
!print '(" corr(y,g_bv):   ",f5.3)',corr1
print '(" corr(s_bv,g_bv):",f5.3)',corr2
!print '(" corr(y,s_bv):   ",f5.3)',corr3
enddo

call cpu_time(z)
Print '(/"Finished in ",f8.2," seconds")',z-z0

CONTAINS

subroutine read_setup
! reads simulation parameters from those produced by program SNP_SIM
open(1,file='SNP_setup')
read(1,*)
read(1,'(6i8)')n_snp,n_qtl,nrec,nprog, nrec
print '(" Read ",i0," records with ",i0," SNP effects")',nrec,n_snp
allocate(snp_val(n_snp),snp_freq(n_snp),genome(2,n_snp))
read(1,*)
read(1,'(10f8.2)')   !var_a,var_e, sim_var
read(1,*)
read(1,'(10f5.2)')snp_freq
read(1,*)
read(1,'(10f8.2)')snp_val
end subroutine


subroutine solve_GSRU
! solves the system of equations by Gauss Seidel Residual Update (Legarra & Misztal, 2008)
integer::i,j
real::conv,conv1,conv_crit=1e-10,lhs,rhs,val,e(nrec)
!
sol=0; e=y

do i=1,nround
   conv=0;conv1=0
   do j=1,neq
      rhs=dot_product(x(:,j),e)+xpx(j)*sol(j)
	  lhs=xpx(j)
	  if (j>1) lhs=lhs+var_ratio  ! equation 1 is mean	
	  val=rhs/lhs
	  conv=conv+val**2;  conv1=conv1+(val-sol(j))**2
	  e=e-x(:,j)*(val-sol(j))
	  sol(j)=val
   enddo
   !print*,i,conv,conv1,sol(1)
   if (conv1/conv<conv_crit) exit
enddo
call cpu_time(z) 
print "('convergence of ',e8.2,' reached in ',i5,' rounds and ',f0.2, ' seconds')",&
       conv1/conv,i,z-z0   
end subroutine

Subroutine solve_pcg
! solves by PCG
!!!   Unfinished !!!

real (r8)::m(neq),r(neq),p(neq),z(neq),w(neq)
real (r8)::alpha,beta,tau,oldtau,conv
real::tt
integer::i
  
m=1/(xpx+var_ratio); m(1)=1/xpx(1)  
sol=0 ; r=xy; 
   
   do i=1,nround
      z=m*r
      tau=dot_product(z,r)
      if (i == 1) then
         beta=0; p=z
       else
         beta=tau/oldtau
         p=z+beta*p
      end if
      w=LHS_times_b(p)  !mult(a,p)
      alpha=tau/dot_product(p,w)
      sol=sol+alpha*p
      if (mod(i,100) /= 0) then
           r=r-alpha*w
         else
           r=xy-LHS_times_b(real(sol,r8))    !mult(a,real(sol,r8))
      endif
      conv=dot_product(r,r)/dot_product(xy,xy)
      if (conv < 1e-14) exit
      oldtau=tau
   enddo
   call cpu_time(tt) 
print "('convergence of ',e8.2,' reached in ',i5,' rounds and ',f0.2, ' seconds')",&
       conv,i,tt-z0  
end subroutine


function LHS_times_b(b) result(a)
! Calculates  a= LHS * b ; for use by solve_pcg
real (r8)::b(:),a(size(b))
integer::i
!
a=0
do i=1,nrec
   a=a+x(i,:)*dot_product(x(i,:),b)
enddo

a(2:)=a(2:)+var_ratio*b(2:)
end function


function outer_prod(x,y) result(z)
!z=ww'
real::x(:),y(:),z(size(x),size(y))
integer::i,j
z=0
do i=1,size(x)
   do j=1,size(y)
      z(i,j)=x(i)*y(j)
   enddo
enddo
end function
   
end program
