Program simul_SNP
! a simple program to simulate SNP genotypes of animals + records
!Ignacy Misztal, Sept 08 - Oct 8,2008

implicit none

integer::   n_snp,& 	!number of SNP
            n_qtl,&		!number of SNP with nonzero effect
		    nprog,&		!number of progeny; 0 if own records generated
		    nrec		!number of records
			
	       ! SNP freq ~ UN(snp_freql, snp_freq_h)		   
real::	  snp_freq_l, snp_freq_h=0.5, &	!snp_freq_h  <= 0.5	   
   		  snp_range ,& ! SNP relative values ~ exp(UN(0,(n_snp-1)/n_snp)))
          var_a,&	!additive variance
		  h2,&     !heritability
		  mean   	!mean

integer:: seed	!different seed generates different sequence of random numbers

integer::i,j,k
real::var_e,sim_var,bv,y,z			  
real,allocatable:: snp_val(:),snp_freq(:)
integer,allocatable::genome(:,:)
integer::times(8)

print '("Program SNP_SIM 1.2(I. Misztal, UGA)")'


call read_par
allocate(snp_val(n_snp),snp_freq(n_snp),genome(2,n_snp))
snp_val=0

!randomize random number generator
call date_and_time(values=times)
do i=1,times(8)
   call random_number(z)
enddo 

! generate gene frequencies and SNP values
do i=1,n_snp
   snp_freq(i)=gen_un(snp_freq_l,snp_freq_h)
enddo

if (snp_range <= 1.0) then
      snp_val(1)=1
   else  
      do i=0,n_QTL-1
         z=exp( real(i)/(n_qtl-1) * log(snp_range)) 
		 if (gen_un(0.,2.)>1) z=-z
         snp_val(n_qtl-i)=z
      enddo
endif
! find simulated genetic variance; to be used for adjustments
sim_var=sum(2*snp_freq*(1-snp_freq)*snp_val**2)

! adjust for genetic variance
snp_val=snp_val*sqrt(var_a/sim_var)

!residual variance as a function of heritability and genetic variance
var_e=var_a*(1-h2)/h2

call store_setup


! generate genome for all SNP
open(2,file='SNP_data')		! generates phenotypes + genotypes
do k=1,2*nrec
   genome=1
   do i=1,n_snp
      do j=1,2
	     z=gen_un(0.,1.0)
   	     if (z < snp_freq(i)) genome(j,i)=2
      enddo
   enddo	

   bv=sum( (genome(1,:)+genome(2,:)-3) *snp_val)

   if (nprog==0) then
	  y=mean+gen_norm(bv,var_e)	! animals own record
      else
          y= mean+bv+gen_norm(0.,(3*var_a+4*var_e)/nprog)	! daughter yield devation 
   endif

   write(2,'(i6,1x,2f8.2,1x,20000i1)')k,y,bv,genome
enddo 

print '("Simulated ",i0," records with ",i0," SNP effects")',nrec,n_snp
print '("Simulated parameters are in file SNP-setup")'
print '("Simulated records are in file SNP-data")'

CONTAINS
	  
function gen_un(a,b) result(x)
! x~un(a,b)
! uses system pseudo-random number generator
!
integer,parameter::n=10
real::a,b,x,y
call random_number(y)
x=a+(b-a)*y
end function


function gen_norm(m,v) result(x)
!x~n(m,v)
real::m,v,x,z(12)
!
call random_number(z)
x=sum(z)-6	!z~un(0,1)
x=m+sqrt(v)*x
end function

subroutine set_random_seed(s)
!changes sequence of pseudo-random numbers
INTEGER :: s,i,n
integer,allocatable::seed(:)
!
CALL RANDOM_SEED(size = n)
ALLOCATE(seed(n))          
seed=(/(i,i=1,n)/)+s
CALL RANDOM_SEED(PUT = seed)         
DEALLOCATE(seed)
end subroutine


subroutine read_par
! reads parameters
character (50)::name
!
!print*,'name of parameter file?'
!read '(a)',name
name='par_snpsim'
print*,'using parameters in file:',name
open(3,file=name)
!
read(3,*); read(3,*)n_snp
print*,n_snp,' SNP'
read(3,*); read(3,*)n_qtl
print*,n_qtl,' of SNP have nozero effects'
read(3,*); read(3,*)nprog
if (nprog==0) then
   print*,'animals with own record'
  else
   print*,nprog,' progenies per sire'
endif
read(3,*); read(3,*)nrec
print*,nrec," records"

read(3,*); read(3,*)snp_freq_l, snp_freq_h
print*,'Minor SNP frequencies range: ',snp_freq_l, snp_freq_h
read(3,*); read(3,*)snp_range
print*,"SNP range of values:",snp_range 
read(3,*); read(3,*)var_a
print*,"Additive variance=",var_a
read(3,*); read(3,*)h2
print*,'h2=',h2
read(3,*); read(3,*)mean
print*,'Mean=',mean
read(3,*); read(3,*)seed
print*,"Random seed=",seed
end subroutine

subroutine store_setup
! stores simulation parameters to be used by subsequent programs
open(1,file='SNP_setup')
write(1,*)'n_snp  n_qtl nrec nprog nrec'
write(1,'(6i8)')n_snp,n_qtl,nrec,nprog, nrec
write(1,*)'var_a,var_e, sim_var '
write(1,'(10f10.2)')var_a,var_e, sim_var
write(1,*)'snp_freq'
write(1,'(10f5.2)')snp_freq
write(1,*)'snp_val'
write(1,'(10f8.2)')snp_val
end subroutine


end program
